unit PrintDrv;
{
	Created by: James "Woody" Woodard
  						2/24/99

  The origins of this component came from the Inprise site and contained most
  of the functionality. I don't remember who the author was but thank him for
  the insight and the main procedures for printing and calculating printer
  metrics. The rest of the implementation is my own and some of the original
  procedures were changed to reflect the addition of the preview mode.

	This component is used as a reporting tool when standard report generators
  don't allow you the flexibility needed for more complex printing. You can
  use this component to provide a more robust method of generating reports.
  Though the amount of work involved in creating a report requires more coding
  than using a report generator, the amount of control you have far outweighs
  the hassles, in my opinion.

  I built this component because I have written several programs that needed to
  be flexible in not only what was printed but where depending on certain criteria
  selectable by the user. Also, most report generators don't allow you to expand
  or contract sections unless you use grids, etc.

  The way this component works is this:
  	Initially, when you issue all of the print commands from start to finish, they
    are stored in an object called Pages. This stores all print commands in lists,
    one list per page. This list of pages and commands is then replayed either to
    the printer or to the preview screen. By capturing the commands this way, it is
    possible for the user to select a printing range of pages instead of the complete
    job if desired.

  Plans for the future:

  	1. Implement a "Cancel" button on the printer progress bar form to allow user to
    cancel printing in progress.

    2. Ability to change page orientation during a report. At present, page orientation
    must be set initially and applies to the total report. As far as I know, in order
    to implement this, a driver file would need to be created that contains the
    proper commands to send to the printer in escape codes.

    3. Ability to pass control strings (escape codes)to the printer. Have to be careful
    about this one because of the preview mode.

    (Any other suggestions would be greatly appreciated)

}
interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, Printers, PrintDr1, PrintDr2;

const
  HeaderLines = 10;                        { Number of allowable header lines }
  FooterLines = 10;                        { Number of allowable footer lines }
  Columns = 20;                            { Number of allowable columns }

type
  THeaderRecord = record
  	Text: string[240];                   { Header text }
    YPosition: single;                   { Inches from the top }
    Alignment: integer;                  { 0:Left 1:Center 2:Right }
    FontName: string[80];                { Font name }
    FontSize: integer;                   { Font size }
    FontStyle: TFontStyles;              { Font style }
  end;

  TFooterRecord = record
    Text: string[240];                   { Footer text }
    YPosition: single;                   { Inches from the top }
    Alignment: integer;                  { 0:Left 1:Center 2:Right }
    FontName: string[80];                { Font name }
    FontSize: integer;                   { Font size }
    FontStyle: TFontStyles;              { Font style }
  end;

  THeaderCoordinates = record
  	XTop: single;
    YTop: single;
    XBottom: single;
    YBottom: single;
    Boxed: boolean;
    Shading: word;
    LineWidth: word;
  end;

  TFooterCoordinates = record
    XTop: single;
    YTop: single;
    XBottom: single;
    YBottom: single;
    Boxed: boolean;
    Shading: word;
    LineWidth: word;
  end;

  TPageNumberRecord = record
    YPosition: single;
    Text: string[240];
    Alignment: word;
    FontName: string[80];
    FontSize: word;
    FontStyle: TFontStyles;
    ShowTotalPages: boolean;
  end;

  TColumnInformationRecord = record
    XPosition: single;
    Length: single;
  end;

  Percentages = (P1_Fifty,P2_Sixty,P3_Seventy,P4_Eighty,P5_Ninety,P6_Full);

  TPrintObject = class(TComponent)
  private
  	TopMargin: integer;               { Top margin in pixels }
    BottomMargin: integer;            { Bottom margin in pixels }
    LeftMargin: integer;              { Left margin in pixels }
    RightMargin: integer;             { Right margin in pixels }
    PixelsPerInchVertical: integer;   { Number of pixels per inch
    																		along Y axis }
    PixelsPerInchHorizontal: integer; { Number of pixels per inch
                                        along X axis }
    TotalPageWidthPixels: integer;    { Full width of page in pixels
                                        includes gutters }
    TotalPageHeightPixels: integer;   { Full height of page in pixels
                                        includes gutters }
    TotalPageHeightInches: single;    { Height of page in inches }
    TotalPageWidthInches: single;     { Width of page in inches }
    GutterLeft: integer;              { Unprintable area on left }
    GutterRight: integer;             { Unprintable area on right }
    GutterTop: integer;               { Unprintable area on top }
    GutterBottom: integer;            { Unprintable area on bottom }
    DetailTop: single;                { Inches from the top where the
                                        detail section starts }
    DetailBottom: single;             { Inches from the top where the
                                        detail section ends }
    LastYPosition: single;            { The Y position where the last
                                        write occurred }
    AutoPaging: boolean;              { Are new pages automatically generated? }
    CurrentTab: single;               { The value of the current tab }
    CurrentFontName: string[30];
    CurrentFontSize: integer;
    CurrentFontStyle: TFontStyles;
    TextMetrics: TTextMetric;
    Header: Array[1..HeaderLines] of THeaderRecord;
    Footer: Array[1..FooterLines] of TFooterRecord;
    ColumnInformation: Array[1..Columns] of TColumnInformationRecord;
    PageNumber: TPageNumberRecord;
    HeaderCoordinates: THeaderCoordinates;
    FooterCoordinates: TFooterCoordinates;
		ThePrinter: TImage; {Used for current canvas during print preview mode}
    TheCanvas: TCanvas; {This is set to current drawing canvas}
    DryRun,	{used for counting total pages before actually printing report}
    PreviewMode: boolean; {Used to set internal preview mode to build pages}
    ActualPreview: boolean; {Used to control users choice of preview mode}
    ScaleByValue: single; {Used to scale the pixels per inch for scaling}
    UseProgress: boolean; {Shows progress bar if set to true}
    ProgressBar: TPrintBarForm; {Progress bar form}
    DocumentTitle: string; {Title to show in print manager for current job}
    PageOrientation: TPrinterOrientation; {Keep track of orientation}
    ShowDialog: boolean; {True to allow print dialog box to show}
    PageNumbering: boolean; {True if page numbering is desired}
    ProgressMax: word; {Sets the progress bar maximum}
    ProgressInc: word; {Sets the progress bar increment}
    PrintTransparent: boolean; {True to print transparent text for backgrounds}
    FPageChange: TNotifyEvent; {Event for user to put any specific printing to
    														be done for every page.}
    PreviewPercent: Percentages;
    procedure OnPageChange;
    procedure SetPreviewMode(Mode: boolean); {set preview or print mode}
    procedure ShowPreview; {call ShowPreview to see the actual report on screen}
    procedure SetShowTotalPages(Mode: boolean); {set whether to display page number
    																						 with total pages (ex. 1 of 26)}
    function GetShowTotalPages: boolean; {returns current setting of show total pages}
		function CalculateLineHeight: integer;
    procedure CalculateMeasurements;
    procedure _DrawBox( XTop,YTop,XBottom,YBottom:word;LineWidth:word; Shading:word );
    procedure SetAutoPaging( Value: boolean );
    procedure Cleanup;

	public
		constructor Create(AOwner: TComponent); override;{need to initialize some default values}
    destructor Destroy; override;
    function GetColumnLeft(Number:Word): single; {returns start in inches of column}
    function GetColumnRight(Number:Word): single; {returns end in inches of column}
    procedure PrintGraphic(X,Y,W,H: single;thePicture: TImage); {Use this procedure to place an image on the printed page.}
    function GetScale: single; {returns the current scale factor}
    procedure SetScale(ScaleFactor: single); {sets the current scale factor}
    procedure UpdateBar(x: integer); {used to set progress bar value}
    procedure IncBar(x: integer); {used to increment the progress bar by X}
    procedure Increment; {used to increment the progress bar by ProgressBarInc value}
    procedure DrawTextBox( XTop,YTop,XBottom,YBottom:single; LineWidth:word;
    											 Shading:word; Text: string; Column: integer;
                           Alignment: word ); {draws a box around written text}
    procedure SetPrnPrinter(Current: TImage); {used when printing to TImage}
    procedure PrintPage(Page: TMWPageList;iPage: integer); {procedure to
        	loop through the current page commands and output to screen or printer}
    procedure PrintFromPreview; {used to print from the preview screen}
		procedure Start; {starts the print job and initializes things}
    procedure Quit; {must be called to complete the print job}
    procedure Abort; {called to abort the print job}
    function InchesToPixelsHorizontal( Inches: single ): integer;
    function InchesToPixelsVertical( Inches: single ): integer;
    function PixelsToInchesHorizontal( Pixels: integer ): single;
    function PixelsToInchesVertical( Pixels: integer ): single;
    function LinesToPixels( Line:integer ): integer;
    procedure SetMargins( Top,Bottom,Left,Right:single );
    procedure SetFontInformation( Name:string; Size:word; Style: TFontStyles );
    procedure WriteLine( X:single; Y:single; Text:string );
    procedure WriteLineRight( Y:single; Text:string );
    procedure WriteLineCenter( Y:single; Text:string );
    procedure WriteLineColumnRight( ColumnNumber:word; Y:single; Text:string );
    procedure WriteLineColumnCenter( ColumnNumber:word; Y:single; Text:string );
    procedure DrawLine( TopX,TopY,BottomX,BottomY:single; LineWidth:word );
    procedure SetLineWidth( Width:word );
    function  GetLineWidth: word;
    procedure SetTab( Inches:single );
    procedure NewPage;
    function  GetLinesPerPage: integer;
    procedure GetPixelsPerInch( var X:word; var Y:word );
    procedure GetPixelsPerPage( var X:word; var Y:word );
    procedure GetGutter( var Top:word; var Bottom:word; var Left:word;
                         var Right:word );
    function  GetTextWidth( Text:string ): integer;
    function  GetLineHeightPixels: word;
    function  GetLineHeightInches: single;
    function  GetPageNumber:integer;
    function  GetColumnsPerLine: integer;
    procedure SetHeaderInformation( Line:integer; YPosition: single;
                                    Text:string; Alignment:word;
                                    FontName:string; FontSize: word;
                     								FontStyle: TFontStyles );
    procedure SetFooterInformation( Line:integer; YPosition: single; Text:string;
                                    Alignment:word; FontName:string; FontSize: word;
                     								FontStyle: TFontStyles );
    procedure WriteHeader;
    procedure WriteFooter;
    procedure SaveCurrentFont;
    procedure RestoreCurrentFont;
    procedure SetDetailTopBottom( Top: single; Bottom: single );
    procedure SetPageNumberInformation( YPosition:single; Text:string;
                   											Alignment:word; IncludeTotal: boolean;
                                        FontName:string; FontSize:word;
                                        FontStyle:TFontStyles );
    procedure WritePageNumber;
    procedure WriteLineColumn( ColumnNumber:word; Y:single; Text:string );
    procedure DrawBox( XTop,YTop,XBottom,YBottom:single; LineWidth:word );
    procedure DrawBoxShaded( XTop,YTop,XBottom,YBottom:single; LineWidth:word;
    												 Shading:word );
    procedure SetHeaderDimensions( XTop,YTop,XBottom,YBottom:single; Boxed: boolean;
    															 LineWidth:word; Shading:word );
    procedure SetFooterDimensions( XTop,YTop,XBottom,YBottom:single; Boxed: boolean;
    															 LineWidth:word; Shading:word );
    procedure CreateColumn( Number:word; XPosition:single; Length:single );
    procedure SetYPosition( YPosition:single );
    function  GetYPosition: single;
    procedure NextLine;
    function  GetLinesLeft: word;
    function  GetLinesInDetailArea: word;
    procedure SetTopOfPage;
    procedure NewLines( Number:word );
    function GetFontName: string;
    function GetFontSize: word;

{Added Published properties for ease of setting certain properties at design time.}
	published
  	{Properties:
    	AutomaticPaging: can be used to automatically do NewPage when printing.
      Title: this text becomes the title of the print job in printer manager
      Preview: whether to print directly to printer of show preview of job
      ShowProgress: display a progress bar or not
      ShowPrintDialog: whether to show print dialog to allow user to select
      								 printer, pages to print, etc.
      Orientation: set page orientation (can only be set for all pages)
      ShowTotalPages: sets whether to display page number as 1 of x or not
      ProgressBarMax: sets the maximum progress bar value
      ProgressBarInc: sets the increment for the progress bar
      Transparent: if you want a background image, this sets the text to print
      						 transparently so it shows through
    }
  	property AutomaticPaging: boolean read AutoPaging write AutoPaging;
  	property Title: string read DocumentTitle write DocumentTitle;
  	property Preview: boolean read ActualPreview write ActualPreview;
    property ShowProgress: boolean read UseProgress write UseProgress;
    property ShowPrintDialog: boolean read ShowDialog write ShowDialog;
    property Orientation: TPrinterOrientation read PageOrientation write PageOrientation;
    property ShowTotalPages: boolean read GetShowTotalPages write SetShowTotalPages;
    property PrintPageNumbers: boolean read PageNumbering write PageNumbering;
    property ProgressBarMax: word read ProgressMax write ProgressMax;
    property ProgressBarInc: word read ProgressInc write ProgressInc;
    property Transparent: boolean read PrintTransparent write PrintTransparent;
    property StartPercent: Percentages read PreviewPercent write PreviewPercent;
		{Events:
      OnNewPage: event for any printing to occur for every page
      	(I use this to put a bitmap on every page like a company logo)
    }
    property OnNewPage: TNotifyEvent read FPageChange write FPageChange;
	end;

procedure Register;

implementation

constructor TPrintObject.Create(AOwner: TComponent);
{Create TPrintObject and set default values}
begin
	inherited Create(AOwner);
  PreviewMode := true;
  ActualPreview := false;
  ScaleByValue := 1.0;
  DocumentTitle := '';
  AutomaticPaging := false;
  ShowProgress := true;
  ShowDialog := true;
  PageOrientation := poPortrait;
  ProgressMax := 100;
  ProgressInc := 1;
  PageNumbering := false;
  PrintTransparent := false;
  PageNumber.ShowTotalPages := true;
  PreviewPercent := P1_Fifty;
end;

destructor TPrintObject.Destroy;
{ Destroy TPrintObject and call TComponent destructor}
begin
  inherited Destroy;
end;

procedure TPrintObject._DrawBox( XTop,YTop,XBottom,YBottom:word;
            										 LineWidth:word; Shading:word );
   { The low level routine which actually draws the box and shades it as
     desired. The paramaters are in pixels and not inches. }
var
	tmpColor: TColor;
  tmpWidth: integer;

begin
	tmpWidth := TheCanvas.Pen.Width;
	TheCanvas.Pen.Width := LineWidth;
  tmpColor := TheCanvas.Brush.Color;
  TheCanvas.Brush.Color := RGB( Shading,Shading,Shading );
	if Not DryRun then
  	TheCanvas.Rectangle( XTop,YTop,XBottom,YBottom );
  TheCanvas.Pen.Width := tmpWidth;
  TheCanvas.Brush.Color := tmpColor;
end;

procedure TPrintObject.Abort;
{Used to abort the print job in progress}
begin
	if PreviewMode then
   	PreviewForm.Abort
  else
   	Printer.Abort;
  Pages.Free;
  Cleanup;
end;

function TPrintObject.CalculateLineHeight: integer;
   { Calculate the height of a line plus the normal amount of space between
     each line }
begin
	Result := TextMetrics.tmHeight + TextMetrics.tmExternalLeading;
end;

function TPrintObject.GetScale: single;
{Returns to current scale multiplier}
begin
	result := ScaleByValue;
end;

procedure TPrintObject.SetScale(ScaleFactor: single);
{Sets the current scale multiplier}
begin
	ScaleByValue := ScaleFactor;
  CalculateMeasurements;
end;

procedure TPrintObject.CalculateMeasurements;
   { Calculate some necessary measurements.  Thanks to Robert Fabiszak
     CompuServe: 70304,2047 for the Escape() Windows API calls. }
var
	pt: TPoint;

begin
   { Call the Windows API function GetTextMetrics() to get the specifics
     of the particular font. }
	GetTextMetrics( TheCanvas.Handle,TextMetrics );

   { Calculate the number of pixels per inch vertical and horizontal.
     'GetDeviceCaps' is a Windows API call. }
  PixelsPerInchVertical := GetDeviceCaps( TheCanvas.Handle,LOGPIXELSY );
  PixelsPerInchHorizontal := GetDeviceCaps( TheCanvas.Handle,LOGPIXELSX );

   { Get the gutter on the left and top.  'Escape' is a Windows API
     call. }
  Escape( TheCanvas.Handle,GETPRINTINGOFFSET,0,Nil,@pt );
  if PreviewMode then
  begin
   	GutterLeft := Trunc(0.25 * PixelsPerInchHorizontal * ScaleByValue);
    GutterTop := Trunc(0.25 * PixelsPerInchVertical * ScaleByValue);
  end
  else
  begin
   	GutterLeft := pt.X;
   	GutterTop := pt.Y;
  end;
  Escape( TheCanvas.Handle,GETPHYSPAGESIZE,0,Nil,@pt );
   {If printing to a TImage, the size doesn't come back as expected so we get the
    size of the preview image}
  if PreviewMode then
  begin
  	pt.X := PreviewForm.GetWidth;//ThePrinter.Width;
    pt.Y := PreviewForm.GetHeight;//ThePrinter.Height;
  end;
  TotalPageWidthPixels := pt.X;
  TotalPageHeightPixels := pt.Y;
  TotalPageWidthInches := pt.X / PixelsPerInchHorizontal;
  TotalPageHeightInches := pt.Y / PixelsPerInchVertical;
   {Need to set these using the scaling value if in preview mode}
  if PreviewMode then
  begin
   	GutterRight := Trunc(0.25 * PixelsPerInchHorizontal * ScaleByValue);
    GutterBottom := Trunc(0.25 * PixelsPerInchVertical * ScaleByValue);
  end
  else
  begin
   	GutterRight := TotalPageWidthPixels - GutterLeft - Printer.PageWidth;
   	GutterBottom := TotalPageHeightPixels - GutterTop - Printer.PageHeight;
  end;
  if ( TopMargin < GutterTop ) then
  	TopMargin := GutterTop;
  if ( BottomMargin < GutterBottom ) then
  	BottomMargin := GutterBottom;
  if ( LeftMargin < GutterLeft ) then
  	LeftMargin := GutterLeft;
  if ( RightMargin < GutterRight ) then
  	RightMargin := GutterRight;
end;

procedure TPrintObject.CreateColumn( Number:word; XPosition:single; Length:single );
{Stores the column information. There is a maximum of 20 columns}
begin
	if (Number > Columns) or (Number < 1) then
  begin
  	ShowMessage('Column value must be between 1 and ' + IntToStr(Columns) + '.');
    exit;
  end;
	if DryRun then
  	Pages.StoreColumn(Number,XPosition,Length);
	ColumnInformation[Number].XPosition := XPosition;
  ColumnInformation[Number].Length := Length;
end;

procedure TPrintObject.DrawBox( XTop,YTop,XBottom,YBottom:single;
																LineWidth:word );
   { Draw a box at the X,Y coordinates passed in the parameters }
var
	BLinePixels,BColPixels,ELinePixels,EColPixels: integer;
  oldStyle: TBrushStyle;

begin
	if DryRun then
    Pages.StoreBox(XTop,YTop,XBottom,YBottom,LineWidth,255);
	if Not PreviewMode then
  begin
   	BLinePixels := InchesToPixelsVertical( YTop ) - GutterTop;
   	ELinePixels := InchesToPixelsVertical( YBottom ) - GutterTop;
   	BColPixels := InchesToPixelsHorizontal( XTop ) - GutterLeft;
   	EColPixels := InchesToPixelsHorizontal( XBottom ) - GutterLeft;
  end
  else
	begin
   	BLinePixels := InchesToPixelsVertical( YTop);
   	ELinePixels := InchesToPixelsVertical( YBottom);
   	BColPixels := InchesToPixelsHorizontal( XTop);
   	EColPixels := InchesToPixelsHorizontal( XBottom);
  end;
  oldStyle := TheCanvas.Brush.Style;
  TheCanvas.Brush.Style := bsClear;
  _DrawBox( BColPixels,BLinePixels,EColPixels,ELinePixels,LineWidth,255 );
  TheCanvas.Brush.Style := oldStyle;
end;

procedure TPrintObject.DrawBoxShaded( XTop,YTop,XBottom,YBottom:single;
																			LineWidth:word; Shading:word );
   { Draw a box at the X,Y coordinates passed in the parameters }
var
	BLinePixels,BColPixels,ELinePixels,EColPixels: integer;
  oldX,oldY,oldWidth: integer;

begin
	if DryRun then
    Pages.StoreBox(XTop,YTop,XBottom,YBottom,LineWidth,Shading);
	if not PreviewMode then
  begin
   	BLinePixels := InchesToPixelsVertical( YTop ) - GutterTop;
   	ELinePixels := InchesToPixelsVertical( YBottom ) - GutterTop;
   	BColPixels := InchesToPixelsHorizontal( XTop ) - GutterLeft;
   	EColPixels := InchesToPixelsHorizontal( XBottom ) - GutterLeft;
  end
  else
  begin
   	BLinePixels := InchesToPixelsVertical( YTop);
   	ELinePixels := InchesToPixelsVertical( YBottom);
   	BColPixels := InchesToPixelsHorizontal( XTop);
   	EColPixels := InchesToPixelsHorizontal( XBottom);
  end;
  _DrawBox( BColPixels,BLinePixels,EColPixels,ELinePixels,1,Shading );

	{Have found that if linewidth is greater than 1, the shading for the
   box does not appear. Have added this in the meantime but need to trace
   down why it happens.}
  with TheCanvas do
  begin
  	oldWidth := Pen.Width;
    oldX := PenPos.X;
    oldY := PenPos.Y;
    Pen.Width := LineWidth;
    MoveTo(BColPixels,BLinePixels);
    LineTo(EColPixels,BLinePixels);
    LineTo(EColPixels,ELinePixels);
    LineTo(BColPixels,ELinePixels);
    LineTo(BColPixels,BLinePixels);
    MoveTo(oldX,oldY);
    Pen.Width := oldWidth;
  end;
end;

procedure TPrintObject.DrawLine( TopX,TopY,BottomX,BottomY:single; LineWidth:word );
   { Draw a line beginning at a particular X,Y coordinate and ending at a
     particular X,Y coordinate. }
var
	TopXPixels, BottomXPixels, TopYPixels, BottomYPixels: integer;
  oldWidth: integer;

begin
	if DryRun then
  	Pages.StoreLine(TopX,TopY,BottomX,BottomY,LineWidth);
	TopXPixels := InchesToPixelsHorizontal( TopX );
  BottomXPixels := InchesToPixelsHorizontal( BottomX );
  TopYPixels := InchesToPixelsVertical( TopY );
  BottomYPixels := InchesToPixelsVertical( BottomY );
  if not PreviewMode then
  begin
   	Dec( TopXPixels,GutterLeft );
   	Dec( BottomXPixels,GutterLeft );
   	Dec( TopYPixels,GutterTop );
   	Dec( BottomYPixels,GutterTop );
  end;
  oldWidth := TheCanvas.Pen.Width;
  TheCanvas.Pen.Width := LineWidth;
  TheCanvas.MoveTo( TopXPixels,TopYPixels );
  if not DryRun then
   		TheCanvas.LineTo( BottomXPixels,BottomYPixels );
  TheCanvas.Pen.Width := oldWidth;
end;

procedure TPrintObject.DrawTextBox( XTop,YTop,XBottom,YBottom:single;
																		LineWidth:word; Shading:word; Text: string;
                                    Column: integer; Alignment: word );
{Routine to draw a shaded box with text printed inside. Added this to make it easier
 than drawing the box and then writing the text.}
var
	BLinePixels,BColPixels,ELinePixels,EColPixels: integer;
  ColorTmp: TColor;
  WidthTmp: integer;

begin
	if DryRun then
    Pages.StoreBox(XTop,YTop,XBottom,YBottom,LineWidth,Shading);
	SaveCurrentFont;
  SetFontInformation(GetFontName,GetFontSize,[]);
  if not PreviewMode then
  begin
   	BLinePixels := InchesToPixelsVertical( YTop ) - GutterTop;
   	ELinePixels := InchesToPixelsVertical( YBottom ) - GutterTop;
   	BColPixels := InchesToPixelsHorizontal( XTop ) - GutterLeft;
   	EColPixels := InchesToPixelsHorizontal( XBottom ) - GutterLeft;
  end
  else
  begin
   	BLinePixels := InchesToPixelsVertical( YTop);
   	ELinePixels := InchesToPixelsVertical( YBottom);
   	BColPixels := InchesToPixelsHorizontal( XTop);
   	EColPixels := InchesToPixelsHorizontal( XBottom);
  end;
  WidthTmp := TheCanvas.Pen.Width;
  _DrawBox( BColPixels,BLinePixels,EColPixels,ELinePixels,LineWidth,Shading );
  RestoreCurrentFont;
  ColorTmp := TheCanvas.Brush.Color;
  TheCanvas.Brush.Color := RGB(Shading,Shading,Shading);
  case Alignment of
  	0: WriteLineColumn(Column,-2,Text);
   	1: WriteLineColumnCenter(Column,-2,Text);
   	2: WriteLineColumnRight(Column,-2,Text);
  end;
  TheCanvas.Brush.Color := ColorTmp;
  TheCanvas.Pen.Width := WidthTmp;
end;

function TPrintObject.GetColumnLeft( Number:word): single;
{Function to return the left X coordinate of the desired column.}
begin
	Result := ColumnInformation[Number].XPosition;
end;

function TPrintObject.GetColumnRight( Number:word): single;
{Function to return the Right X coordinate of the desired column.}
begin
	Result := ColumnInformation[Number].XPosition + ColumnInformation[Number].Length;
end;

function TPrintObject.GetColumnsPerLine: integer;
   { How many columns are there in a Line? }
var
	Pixels: integer;

begin
	Pixels := TotalPageWidthPixels - GutterLeft - GutterRight;
  Result := Pixels div TheCanvas.TextWidth( 'B' );
end;

function TPrintObject.GetFontName: string;
   { Return the current font name }
begin
	Result := TheCanvas.Font.Name;
end;

function TPrintObject.GetFontSize: word;
   { Return the current font size }
begin
	Result := TheCanvas.Font.Size;
end;

procedure TPrintObject.GetGutter( var Top:word; var Bottom:word; var Left:word;
            											var Right:word );
{ Returns the gutter (non-printable) setting of the page }
begin
	Top := GutterTop;
  Bottom := GutterBottom;
  Left := GutterLeft;
  Right := GutterRight;
end;

function TPrintObject.GetLineHeightInches: single;
{ Returns the height of current line in inches }
begin
	Result := PixelsToInchesVertical( GetLineHeightPixels );
end;

function TPrintObject.GetLineHeightPixels: word;
{ Returns the height of current line in pixels }
begin
	Result := CalculateLineHeight;
end;

function TPrintObject.GetLinesInDetailArea: word;
   { Return the number of lines in the detail area }
begin
  Result := InchesToPixelsVertical( DetailBottom - DetailTop ) div CalculateLineHeight;
end;

function TPrintObject.GetLinesLeft: word;
   { Return the number of lines left in the detail area }
var
	Lines: single;

begin
	Lines := (DetailBottom - LastYPosition) / GetLineHeightInches;
  Result := Trunc(Lines);
end;

function TPrintObject.GetLinesPerPage: integer;
   { Return the number of lines on the entire page }
begin
	Result := (TotalPageHeightPixels - GutterTop - GutterBottom) div CalculateLineHeight;
end;

function TPrintObject.GetLineWidth: word;
{ Returns the current drawing width of the pen }
begin
	Result := TheCanvas.Pen.Width;
end;

function TPrintObject.GetPageNumber;
   { Return the current page number }
begin
	if PreviewMode then
   	Result := PreviewForm.GetPageNumber
  else
   	Result := Printer.PageNumber;
end;

procedure TPrintObject.GetPixelsPerInch( var X:word; var Y:word );
{ Returns the current pixels per inch of the canvas }
begin
	X := PixelsPerInchHorizontal;
  Y := PixelsPerInchVertical;
end;

procedure TPrintObject.GetPixelsPerPage( var X:word; var Y:word );
{ Returns the number of pixels across the printable area of the page }
begin
	X := TotalPageWidthPixels - GutterLeft - GutterRight;
  Y := TotalPageHeightPixels - GutterTop - GutterBottom;
end;

function TPrintObject.GetTextWidth( Text:string ): integer;
   { Return the width of the text contained in 'Text' in pixels }
begin
	Result := TheCanvas.TextWidth( Text );
end;

function TPrintObject.GetYPosition: single;
{ Returns the current vertical position of pen on the canvas }
begin
	Result := LastYPosition;
end;

function TPrintObject.InchesToPixelsHorizontal( Inches: single ): integer;
   { Convert the horizontal inches represented in 'Inches' to pixels }
var
	Value: single;

begin
	Value := Inches * PixelsPerInchHorizontal * ScaleByValue;
  Result := Trunc(Value);
end;

function TPrintObject.InchesToPixelsVertical( Inches: single ): integer;
   { Convert the vertical inches represented in 'Inches' to pixels }
var
	Value: single;

begin
	Value := Inches * PixelsPerInchVertical * ScaleByValue;
  Result := Trunc(Value);
end;

function TPrintObject.LinesToPixels( Line:integer ): integer;
   { Calculate the number of vertical pixels in 'Line' }
begin
	if ( Line <= 0 ) then
  	Line := 1;
  Result := (Line-1) * CalculateLineHeight;
end;

procedure TPrintObject.NewLines( Number:word );
   { Generate the number of line feeds represented in 'Number' }
var
	I: word;

begin
	for I := 1 to Number do
  	NextLine;
end;

procedure TPrintObject.NewPage;
   { Issue a new page }
begin
	WriteHeader;
  WriteFooter;
  WritePageNumber;
  OnPageChange;
  LastYPosition := DetailTop - GetLineHeightInches;
  if DryRun then
  begin
  	Pages.NewPage;
    SetFontInformation(CurrentFontName,CurrentFontSize,CurrentFontStyle);
    inc(TotalPages);
  end
  else if not PreviewMode then
  		Printer.NewPage;
end;

procedure TPrintObject.NextLine;
{ Advances pen pos to next line on canvas }
begin
	if DryRun then
  	Pages.StoreGeneral(commNewLine);
	LastYPosition := LastYPosition + GetLineHeightInches;
end;

function TPrintObject.PixelsToInchesHorizontal( Pixels: integer ): single;
{ Converts "Pixels" to inches using horizontal resolution }
begin
	Result := Pixels / PixelsPerInchHorizontal;
end;

function TPrintObject.PixelsToInchesVertical( Pixels: integer ): single;
{ Converts "Pixels" to inches using vertical resolution }
begin
	Result := Pixels / PixelsPerInchVertical;
end;

procedure TPrintObject.PrintGraphic(X,Y,W,H: single;thePicture: TImage);
{ Used to copy contents of TImage to printing canvas }
var
  x1,x2,y1,y2: integer;
  ScreenToPrinter: single;

begin
	if DryRun then
  begin
  	Pages.StoreGraphic(X,Y,H,W,thePicture);
    exit;
  end;
  if PixelsPerInchHorizontal > PixelsPerInchVertical then
  	ScreenToPrinter := PixelsPerInchHorizontal / Screen.PixelsPerInch
  else
  	ScreenToPrinter := PixelsPerInchVertical / Screen.PixelsPerInch;
  if W = -1.0 then
  	x2 := Trunc((thePicture.Width * ScreenToPrinter) * ScaleByValue)
  else
  	x2 := Trunc(InchesToPixelsHorizontal(W) * ScaleByValue);
  W := PixelsToInchesHorizontal(x2) / ScaleByValue;
  if X = -1.0 then
  	if TotalPageWidthPixels > InchesToPixelsHorizontal(W) then
    	x1 := Trunc(((TotalPageWidthPixels * ScaleByValue) - InchesToPixelsHorizontal(W)) / 2)
    else
    	x1 := 0
  else
		x1 := Trunc(InchesToPixelsHorizontal(X) * ScaleByValue);
  if H = -1.0 then
  	y2 := Trunc((thePicture.Height * ScreenToPrinter) * ScaleByValue)
  else
  	y2 := Trunc(InchesToPixelsVertical(H) * ScaleByValue);
  H := PixelsToInchesVertical(y2) / ScaleByValue;
  if Y = -1.0 then
  	if TotalPageHeightPixels > InchesToPixelsVertical(H) then
    	y1 := Trunc(((TotalPageHeightPixels * ScaleByValue) - InchesToPixelsVertical(H)) / 2)
    else
    	y1 := 0
  else
  	y1 := Trunc(InchesToPixelsVertical(Y) * ScaleByValue);
	TheCanvas.StretchDraw(rect(x1,y1,x1+x2,y1+y2),thePicture.Picture.Graphic);
end;

procedure TPrintObject.Quit;
   { 'Quit' must always be called when printing is completed. If preview
   		mode is set, then it displays the preview form. }
begin
	WriteHeader;
  WriteFooter;
  WritePageNumber;
  OnPageChange;
  DryRun := false;
  if UseProgress then
  begin
  	ProgressBar.Close;
  	ProgressBar.Free;
  end;
  Pages.FromPage := 1;
  Pages.ToPage := Pages.Count;
  if ActualPreview then
    ShowPreview
  else
  	PrintFromPreview;
  Pages.Free;
  Cleanup;
end;

procedure TPrintObject.PrintFromPreview;
{ Used to print the actual commands that were stored either from the preview
  form or directly when Quit is called to end the print job. }
var
	x: integer;
  xrInc, xrCurrent: single;
  oldCanvas: TCanvas;

begin
	if ShowDialog then
		if not Pages.SetPrintOptions then
  		exit;
	PreviewMode := false;
  xrInc := 100 / ((Pages.ToPage + 1) - Pages.FromPage);
  xrCurrent := 0.0;
	if UseProgress then
  begin
  	ProgressBar := TPrintBarForm.Create(Application);
  	with ProgressBar do
  	begin
  		SetDisplayText('Printing report............');
  		Left := Application.MainForm.Left + (Application.MainForm.Width - Width) div 2;
  		Top := Application.MainForm.Top + (Application.MainForm.Height - Height) div 2;
  		Show;
  		UpdateBar(0);
  	end;
  end;
  oldCanvas := TheCanvas;
  Printer.Title := DocumentTitle;
	if PreviewMode then
  	PreviewForm.SetOrient(PageOrientation)
  else
  	Printer.Orientation := PageOrientation;
  Printer.BeginDoc;
  CalculateMeasurements;
  TheCanvas := Printer.Canvas;
  for x := Pages.FromPage to Pages.ToPage do
  begin
  	PrintPage(Pages.Get(x),x);
    if x < Pages.ToPage then
    	Printer.NewPage;
    xrCurrent := xrCurrent + xrInc;
    if UseProgress then
    	ProgressBar.UpdateBar(Round(xrCurrent));
  end;
  if UseProgress then
  	ProgressBar.UpdateBar(100);
  Printer.EndDoc;
  PreviewMode := true;
  TheCanvas := oldCanvas;
  if UseProgress then
  begin
  	ProgressBar.Close;
  	ProgressBar.Free;
  end;
end;

procedure TPrintObject.RestoreCurrentFont;
{ Restores the font information after being saved }
begin
	SetFontInformation( CurrentFontName,CurrentFontSize,CurrentFontStyle );
end;

procedure TPrintObject.SaveCurrentFont;
{ Saves the current font information }
begin
	CurrentFontName := TheCanvas.Font.Name;
  CurrentFontSize := TheCanvas.Font.Size;
  CurrentFontStyle := TheCanvas.Font.Style;
end;

procedure TPrintObject.SetAutoPaging( Value: boolean );
{ Controls whether automatic paging is used when reaching the end
  of the detail area of the page }
begin
	AutoPaging := Value;
end;

procedure TPrintObject.SetDetailTopBottom( Top: single; Bottom: single );
{ Sets the top and bottom margins of the detail printing area of the
  report }
begin
	if DryRun then
  	Pages.StoreDetailArea(Top,Bottom);
	DetailTop := Top;
  DetailBottom := Bottom;
  LastYPosition := Top - GetLineHeightInches;
end;

procedure TPrintObject.SetFontInformation( Name:string; Size:word; Style: TFontStyles );
   { Change the current font information }
begin
	if DryRun then
  	Pages.StoreFontChange(Name,Size,Style);
	TheCanvas.Font.Name := Name;
  TheCanvas.Font.Size := Size;
  TheCanvas.Font.Style := Style;
  TheCanvas.Pen.Mode := pmMask;
  CalculateMeasurements;
end;

procedure TPrintObject.SetFooterDimensions( XTop,YTop,XBottom,YBottom:single;
																						Boxed: boolean; LineWidth:word;
                                            Shading:word );
{ Sets the area of the page footer and whether it should be boxed and shaded }
begin
	if DryRun then
  	Pages.StoreFooterDim(XTop,YTop,XBottom,YBottom,Boxed,LineWidth,Shading);
	FooterCoordinates.XTop := XTop;
  FooterCoordinates.XBottom := XBottom;
  FooterCoordinates.YTop := YTop;
  FooterCoordinates.YBottom := YBottom;
  FooterCoordinates.Boxed := Boxed;
  FooterCoordinates.LineWidth := LineWidth;
  FooterCoordinates.Shading := Shading;
end;

procedure TPrintObject.SetFooterInformation( Line:integer; YPosition: single;
            Text:string; Alignment:word; FontName:string; FontSize: word;
            FontStyle: TFontStyles );
{ Sets the footer information to be printed }
begin
	if ( Line > FooterLines ) then
  	Exit;
  if DryRun then
  	Pages.StoreFooterInf(Line,YPosition,Text,Alignment,FontName,FontSize,FontStyle);
  Footer[Line].Text := Text;
  Footer[Line].YPosition := YPosition;
  Footer[Line].Alignment := Alignment;
  Footer[Line].FontName := FontName;
  Footer[Line].FontSize := FontSize;
  Footer[Line].FontStyle := FontStyle;
end;

procedure TPrintObject.SetHeaderDimensions( XTop,YTop,XBottom,YBottom:single;
																						Boxed: boolean; LineWidth:word;
                                            Shading:word );
{ Sets the area of the header for the page }
begin
	if DryRun then
  	Pages.StoreHeaderDim(XTop,YTop,XBottom,YBottom,Boxed,LineWidth,Shading);
	HeaderCoordinates.XTop := XTop;
  HeaderCoordinates.XBottom := XBottom;
  HeaderCoordinates.YTop := YTop;
  HeaderCoordinates.YBottom := YBottom;
  HeaderCoordinates.Boxed := Boxed;
  HeaderCoordinates.LineWidth := LineWidth;
  HeaderCoordinates.Shading := Shading;
end;

procedure TPrintObject.SetHeaderInformation( Line:integer; YPosition: single;
            Text:string; Alignment:word; FontName:string; FontSize: word;
            FontStyle: TFontStyles );
{ Sets the header information to be printed }
begin
	if ( Line > HeaderLines ) then
		Exit;
  if DryRun then
  	Pages.StoreHeaderInf(Line,YPosition,Text,Alignment,FontName,FontSize,FontStyle);
  Header[Line].Text := Text;
  Header[Line].YPosition := YPosition;
  Header[Line].Alignment := Alignment;
  Header[Line].FontName := FontName;
  Header[Line].FontSize := FontSize;
  Header[Line].FontStyle := FontStyle;
end;

procedure TPrintObject.SetLineWidth( Width:word );
{ Sets the current pen width for drawing }
begin
	if DryRun then
  	Pages.StoreLineWidth(Width);
  TheCanvas.Pen.Width := Width;
end;

procedure TPrintObject.SetMargins( Top,Bottom,Left,Right:Single );
   { Set the top, bottom, left and right margins in inches }
begin
   { If the sum of the left and right margins exceeds the width of the page,
     set the left margin to the value of 'GutterLeft' and set the right
     margin to the value of 'GutterRight' }
	if DryRun then
  	Pages.StoreMargins(Top,Bottom,Left,Right);
	if ( Left + Right >= TotalPageWidthInches ) then
  begin
  	Left := PixelsToInchesHorizontal(GutterLeft);
    Right := PixelsToInchesHorizontal(GutterRight);
  end;
  if ( Left <= 0 ) then
  	Left := PixelsToInchesHorizontal(GutterLeft);
  if ( Right <= 0 ) then
  	Right := PixelsToInchesHorizontal(GutterRight);
   { If the sum of the top and bottom margins exceeds the height of the
     page, set the top margin to the value of 'GutterTop' and set the
     bottom margin to the value of 'GutterBottom' }
  if ( Top + Bottom >= TotalPageHeightInches ) then
  begin
  	Top := PixelsToInchesVertical(GutterTop);
    Bottom := PixelsToInchesVertical(GutterBottom);
  end;
  if ( Top <= 0 ) then
  	Top := PixelsToInchesVertical(GutterTop);
  if ( Bottom <= 0 ) then
  	Bottom := PixelsToInchesVertical(GutterBottom);
   { Convert everything to pixels }
  TopMargin := InchesToPixelsVertical( Top );
  if ( TopMargin < GutterTop ) then
  	TopMargin := GutterTop;
  BottomMargin := InchesToPixelsVertical( Bottom );
  if ( BottomMargin < GutterBottom ) then
  	BottomMargin := GutterBottom;
  LeftMargin := InchesToPixelsHorizontal( Left );
  if ( LeftMargin < GutterLeft ) then
  	LeftMargin := GutterLeft;
  RightMargin := InchesToPixelsHorizontal( Right );
  if ( RightMargin < GutterRight ) then
  	RightMargin := GutterRight;
end;

procedure TPrintObject.SetPageNumberInformation( YPosition:single;
   Text:string; Alignment:word; IncludeTotal: boolean;FontName:string;
   FontSize:word; FontStyle:TFontStyles );
{ Sets the manner of printing the page number. The "Text" parameter should be in
  the form of a format string }
begin
	if DryRun then
  	Pages.StorePageNumber(YPosition,Text,Alignment,IncludeTotal,FontName,FontSize,FontStyle);
	PageNumber.Text := Text;
  PageNumber.YPosition := YPosition;
  PageNumber.Alignment := Alignment;
  PageNumber.FontName := FontName;
  PageNumber.FontSize := FontSize;
  PageNumber.FontStyle := FontStyle;
  PageNumber.ShowTotalPages := IncludeTotal;
end;

procedure TPrintObject.SetPreviewMode(Mode: boolean);
{ Sets the preview mode and canvas }
begin
	Case Mode of
  True:
  	begin
      ThePrinter := PreviewForm.GetCurrentPage;
    	TheCanvas := ThePrinter.Canvas;
    end;
  end;
  PreviewMode := Mode;
end;

procedure TPrintObject.SetPrnPrinter(Current: TImage);
{ Sets the current printing canvas of the preview form }
begin
	ThePrinter := Current;
  TheCanvas := ThePrinter.Canvas;
end;

procedure TPrintObject.SetTab( Inches:single );
{ Set a tab location on the page }
begin
 	if DryRun then
  	Pages.StoreTab(Inches);
  CurrentTab := Inches;
end;

procedure TPrintObject.SetTopOfPage;
{ Sets the printing canvas to the top of the detail area on the page }
begin
	if DryRun then
  	Pages.StoreGeneral(commTopOfPage);
  LastYPosition := DetailTop;
end;

procedure TPrintObject.SetYPosition( YPosition:single );
{ Sets the current line position on the printing canvas }
begin
	if DryRun then
  	Pages.StoreSetY(YPosition);
	LastYPosition := YPosition;
end;

procedure TPrintObject.ShowPreview;
{ Displays the preview form }
begin
	If not PreviewMode then
  	exit;
  PreviewForm.StartPercent := ord(PreviewPercent);
  PreviewForm.ShowModal;
end;

procedure TPrintObject.Start;
   { This function MUST be called first before any other printing function }
var
	Top,Bottom,Left,Right: single;
  I: integer;

Begin
{Added this to implement the stored commands procedures. We force the printing to
 run through first in DryRun mode. This gives us the ability to store all print
 commands and then replay them when the Quit command is executed. If the preview
 mode is set, then we need to pass control to the preview form. It will drive the
 page printing through a procedure call hook that we set with OnPrintPage event of
 the preview object.}
// 	UseProgress := UseBar;
  Application.CreateForm(TPreviewForm,PreviewForm);
	SetPreviewMode(true);
  Pages := TMWPages.Create;
  Pages.NewPage;
	DryRun := true;
	TotalPages := 1;
  ScaleByValue := 1.0;
  if UseProgress then
  begin
  	ProgressBar := TPrintBarForm.Create(Application);
    with ProgressBar do
    begin
  		Left := Application.MainForm.Left + (Application.MainForm.Width - Width) div 2;
  		Top := Application.MainForm.Top + (Application.MainForm.Height - Height) div 2;
      SetDisplayText('Preparing report............');
      PrintBar.Max := ProgressMax;
  		Show;
  		UpdateBar(0);
    end;
  end;
	if PreviewMode then
  	PreviewForm.SetOrient(PageOrientation)
  else
  	Printer.Orientation := PageOrientation;
	if Not PreviewMode then {Added if check for preview mode  Woody 2/20/99}
  begin
   	Printer.Title := DocumentTitle;
   	Printer.BeginDoc;
    TheCanvas := Printer.Canvas;
  end
  else
  begin
  	PreviewForm.SetPreviewPrn(self);
   	SetPrnPrinter(PreviewForm.GetCurrentPage);
  end;
  AutoPaging := True;
  CalculateMeasurements;
  PageNumber.Text := '';
  Top := PixelsToInchesVertical( GutterTop );
  Bottom := PixelsToInchesVertical( GutterBottom );
  Left := PixelsToInchesHorizontal( GutterLeft );
  Right := PixelsToInchesHorizontal( GutterRight );
  SetMargins( Top,Bottom,Left,Right );
  for I := 1 to HeaderLines do
  	Header[I].Text := '';
  HeaderCoordinates.Boxed := False;
  HeaderCoordinates.Shading := 0;
  for I := 1 to FooterLines do
  	Footer[I].Text := '';
  FooterCoordinates.Boxed := False;
  FooterCoordinates.Shading := 0;
  CurrentTab := 0.0;
  LastYPosition := 0.0;
end;

procedure TPrintObject.WriteFooter;
   { If any footers are defined, write them }
var
	I: integer;
  Temp: boolean;

begin
	SaveCurrentFont;
   { Set 'AutoPaging' off.  Otherwise the footer will not get written
     correctly. }
  Temp := AutoPaging;
  AutoPaging := False;
  for I := 1 to FooterLines do
  begin
  	if ( Length(Footer[I].Text) > 0 ) then
    begin
    	with Footer[I] do
      begin
      	SetFontInformation( FontName,FontSize,FontStyle );
        if ( Alignment = 0 ) then
        	WriteLine( PixelsToInchesHorizontal(LeftMargin), YPosition, Text );
        if ( Alignment = 1 ) then
        	WriteLineCenter( YPosition, Text );
        if ( Alignment = 2 ) then
        	WriteLineRight( YPosition, Text );
      end;
    end;
  end;
  RestoreCurrentFont;
   { Does the user desire a box around the footer? }
  if ( FooterCoordinates.Boxed = True ) then
  begin
  	if ( FooterCoordinates.Shading > 0 ) then
    	DrawBoxShaded( FooterCoordinates.XTop,FooterCoordinates.YTop,
      	FooterCoordinates.XBottom,FooterCoordinates.YBottom,
      	FooterCoordinates.LineWidth,FooterCoordinates.Shading)
    else
    	DrawBox( FooterCoordinates.XTop,FooterCoordinates.YTop,
      	FooterCoordinates.XBottom,FooterCoordinates.YBottom,
        FooterCoordinates.LineWidth );
  end;
  AutoPaging := Temp;
end;

procedure TPrintObject.WriteHeader;
   { If any headers are defined, write them }
var
	I: Integer;

begin
	SaveCurrentFont;
  for I := 1 to HeaderLines do
  begin
  	if ( Length(Header[I].Text) > 0 ) then
    begin
    	with Header[I] do
      begin
      	SetFontInformation( FontName,FontSize,FontStyle );
        if ( Alignment = 0 ) then
        	WriteLine( PixelsToInchesHorizontal(LeftMargin), YPosition, Text );
        if ( Alignment = 1 ) then
        	WriteLineCenter( YPosition, Text );
        if ( Alignment = 2 ) then
        	WriteLineRight( YPosition, Text );
      end;
    end;
  end;
  RestoreCurrentFont;
   { Does the user desire a box around the header? }
  if ( HeaderCoordinates.Boxed = True ) then
  begin
  	if ( HeaderCoordinates.Shading > 0 ) then
    	DrawBoxShaded( HeaderCoordinates.XTop,HeaderCoordinates.YTop,
      	HeaderCoordinates.XBottom,HeaderCoordinates.YBottom,
        HeaderCoordinates.LineWidth,HeaderCoordinates.Shading)
    Else
    	DrawBox( HeaderCoordinates.XTop,HeaderCoordinates.YTop,
      	HeaderCoordinates.XBottom,HeaderCoordinates.YBottom,
        HeaderCoordinates.LineWidth );
  End;
End;

procedure TPrintObject.WriteLine( X:single; Y:single; Text:string );
   { Write some text.  The parameters represent inches from the left ('X')
     and top ('Y') margins. }
var
	XPixels: integer;
  YPixels: integer;

begin
   { How many pixels are there in the inches represented by 'X'? }
	if DryRun then
  	if (y = -1) then
    	Pages.StoreText(X,LastYPosition + GetLineHeightInches,Text,TheCanvas.Brush.Color)
    else if (y = -2) then
    	Pages.StoreText(X,LastYPosition,Text,TheCanvas.Brush.Color)
    else
  		Pages.StoreText(X,Y,Text,TheCanvas.Brush.Color);
	if ( X >= 0.0 ) then
  	XPixels := InchesToPixelsHorizontal( X )
  else
  	XPixels := LeftMargin;
  if ( XPixels < GutterLeft ) then
  	XPixels := GutterLeft;
   { If there is a tab set, increase 'XPixels' by the amount of the tab }
  if ( CurrentTab > 0.0 ) then
  	Inc( XPixels,InchesToPixelsHorizontal(CurrentTab) );
   { How many pixels are there in the inches represented by 'Y'? }
  YPixels := 0;
  if ( Y > -0.01 ) then
   { Printing will occur at an absolute location from the top of the
     page. }
  begin
  	YPixels := InchesToPixelsVertical( Y );
    if ( YPixels < GutterTop ) then
    	YPixels := GutterTop;
    if ( YPixels > TotalPageHeightPixels ) then
    	YPixels := TotalPageHeightPixels - GutterBottom;
    LastYPosition := Y;
  end;
  if ( Y = -1 ) then
   { Write the text at the next line }
  begin
  	if ( AutoPaging = True ) then
    begin
     { If the next line we're going to write to exceeds beyond the
       bottom of the detail section, issue a new page }
     	if ( LastYPosition + GetLineHeightInches > DetailBottom ) then
      	NewPage;
    end;
    YPixels := InchesToPixelsVertical( LastYPosition + GetLineHeightInches );
    LastYPosition := LastYPosition + GetLineHeightInches;
  end;
  if ( Y = -2 ) then
   { Write the text on the current line }
   	YPixels := InchesToPixelsVertical( LastYPosition );
  if PrintTransparent then
  	TheCanvas.Brush.Style := bsClear
  else
  	TheCanvas.Brush.Style := bsSolid;
  if not DryRun then
  	if PreviewMode then
   		TheCanvas.TextOut( XPixels,YPixels,Text )
    else
   		TheCanvas.TextOut( XPixels-GutterLeft,YPixels-GutterTop,Text );
end;

procedure TPrintObject.WriteLineCenter( Y:single; Text:string );
   { Print a line of text centered at Y inches from the top }
var
	PixelLength: integer;
  StartPixel: integer;

begin
   { How many pixels does the text in 'Text' require? }
	PixelLength := TheCanvas.TextWidth( Text );
   { Calculate where printing should start }
  StartPixel := ((GutterLeft+(TotalPageWidthPixels-GutterRight)) div 2) - (PixelLength div 2);
  SetTab( 0.0 );
  WriteLine( PixelsToInchesHorizontal(StartPixel),Y,Text );
  SetTab( CurrentTab );
end;

procedure TPrintObject.WriteLineColumn( ColumnNumber:word; Y:single; Text:string );
   { Write text, left aligned against the column represented by
     'ColumnInformation[ColumnNumber]' }
begin
	WriteLine( ColumnInformation[ColumnNumber].XPosition,Y,Text );
end;

procedure TPrintObject.WriteLineColumnCenter( ColumnNumber:word; Y:single; Text:string );
   { Print a line of text centered within the column number represented by
     'ColumnNumber', at Y inches from the top }
var
	PixelLength: Integer;
  StartPixel: Integer;

begin
   { How many pixels does the text in 'Text' require? }
	PixelLength := TheCanvas.TextWidth( Text );
   { Calculate where printing should start }
	 { Pixels := InchesToPixelsHorizontal( ColumnInformation[ColumnNumber].Length );}
  StartPixel := (InchesToPixelsHorizontal( ColumnInformation[ColumnNumber].Length )
  	div 2) + InchesToPixelsHorizontal(ColumnInformation[ColumnNumber].XPosition) -
    (PixelLength div 2);
  SetTab( 0.0 );
  WriteLine( PixelsToInchesHorizontal(StartPixel),Y,Text );
  SetTab( CurrentTab );
end;

procedure TPrintObject.WriteLineColumnRight( ColumnNumber:word; Y:single; Text:string );
   { Write text, right aligned against the column represented by
     'ColumnInformation[ColumnNumber]' }
var
	PixelLength: word;
  StartPixel: word;

begin
   { How many pixels does the text in 'Text' require? }
	PixelLength := TheCanvas.TextWidth( Text );
   { Calculate where printing should start }
  StartPixel := InchesToPixelsHorizontal( ColumnInformation[ColumnNumber].XPosition +
  	ColumnInformation[ColumnNumber].Length ) - PixelLength;
  SetTab( 0.0 );
  WriteLine( PixelsToInchesHorizontal(StartPixel),Y,Text );
  SetTab( CurrentTab );
end;

procedure TPrintObject.WriteLineRight( Y:single; Text:string );
   { Print a line of text right justified 'Y' inches from the top }
var
	PixelLength: word;
  StartPixel: word;

begin
   { How many pixels does the text in 'Text' require? }
	PixelLength := TheCanvas.TextWidth( Text );
   { Calculate where printing should start }
  StartPixel := (TotalPageWidthPixels-GutterLeft-GutterRight) - PixelLength;
  SetTab( 0.0 );
  WriteLine( PixelsToInchesHorizontal(StartPixel),Y,Text );
  SetTab( CurrentTab );
end;

procedure TPrintObject.WritePageNumber;
{ Print the current page number on page if page numbering is set }
var
	Buffer: string;
  Temp: boolean;

begin
	if Not PageNumbering  or (PageNumber.Text = '') then
  	exit;
  try
		if PreviewMode then
  		Buffer := Format( PageNumber.Text,[PreviewForm.GetPageNumber] )
  	else
  		Buffer := Format( PageNumber.Text,[Printer.PageNumber] );
  except
  	ShowMessage('Error in page number formatting.');
    exit;
  end;
	if DryRun then
  	Pages.StoreGeneral(commWritePage);
  if PageNumber.ShowTotalPages then
  	Buffer := Buffer + ' of ' + IntToStr(TotalPages);
  SaveCurrentFont;
  SetFontInformation( PageNumber.FontName,PageNumber.FontSize,PageNumber.FontStyle );
  Temp := AutoPaging;
  AutoPaging := False;
  if ( PageNumber.Alignment = 0 ) then
  	WriteLine( PixelsToInchesHorizontal(LeftMargin), PageNumber.YPosition, Buffer );
  if ( PageNumber.Alignment = 1 ) then
  	WriteLineCenter( PageNumber.YPosition, Buffer );
  if ( PageNumber.Alignment = 2 ) then
  	WriteLineRight( PageNumber.YPosition, Buffer );
  AutoPaging := Temp;
  RestoreCurrentFont;
end;

procedure TPrintObject.PrintPage(Page: TMWPageList;iPage: integer);
{ This procedure will loop through a given pages command list and output that
  information to the current printing canvas }
var
	x: integer;
  oldColor: TColor;
  tCommand: TMWPrintCommand;
  Buffer: string;
  SkipNext: integer;

begin
	SkipNext := 0;
  for x := 1 to Page.Count do
  begin
  	if SkipNext = 0 then
    begin
  	tCommand := Page.Get(x);
    case tCommand.Command of
    	commBox:
      	try
      	with tCommand as TMWBox do
          DrawBoxShaded(X1,Y1,X2,Y2,LineWidth,Shading);
        except
        	ShowMessage('Error in Box Drawing command.');
        end;
      commLine:
      	try
      	with tCommand as TMWLine do
        	DrawLine(X1,Y1,X2,Y2,LineWidth);
        except
        	ShowMessage('Error in Line Drawing command.');
        end;
      commNewLine:
      	try
        	NextLine;
        except
        	ShowMessage('Error in Next Line command.');
        end;
      commGraphic:
      	try
      	with tCommand as TMWGraphic do
        	PrintGraphic(X,Y,Height,Width,Bitmap);
        except
        	ShowMessage('Error in Graphics Drawing routine.');
        end;
      commDetailArea:
      	try
      	with tCommand as TMWDetail do
        	SetDetailTopBottom(X,Y);
        except
        	ShowMessage('Error in Setting Detail Top/Bottom settings.');
        end;
      commFont:
      	try
      	with tCommand as TMWFont do
        	SetFontInformation(Name,Size,Style);
        except
        	ShowMessage('Error in Changing Font information');
        end;
      commFooterDim:
      	try
      	with tCommand as TMWFooterDim do
        	SetFooterDimensions(X1,Y1,X2,Y2,Boxed,LineWidth,Shading);
        except
        	ShowMessage('Error in Setting Footer Dimensions.');
        end;
      commFooterInf:
      	try
      	with tCommand as TMWFooterInf do
        	SetFooterInformation(Line,yPosition,Text,Alignment,FontName,Size,Style);
        except
        	ShowMessage('Error in Setting Footer Information.');
        end;
      commHeaderDim:
      	try
      	with tCommand as TMWHeaderDim do
        	SetHeaderDimensions(X1,Y1,X2,Y2,Boxed,LineWidth,Shading);
        except
        	ShowMessage('Error in Setting Header Dimensions.');
        end;
      commHeaderInf:
      	try
      	with tCommand as TMWHeaderInf do
        	SetHeaderInformation(Line,yPosition,Text,Alignment,FontName,Size,Style);
        except
        	ShowMessage('Error in Setting Header Information.');
        end;
      commLineWidth:
      	try
      	with tCommand as TMWLineWidth do
        	SetLineWidth(LineWidth);
        except
        	ShowMessage('Error in Setting Line Width.');
        end;
      commMargins:
      	try
      	with tCommand as TMWMargins do
        	SetMargins(X1,Y1,X2,Y2);
        except
        	ShowMessage('Error in Setting Margins.');
        end;
      commOrient:
      	try
      	with tCommand as TMWOrient do
        	PageOrientation := Orientation;
        except
        	ShowMessage('Error in Setting Page Orientation.');
        end;
      commWritePage:
      	try
          tCommand := Page.Get(x+1);
      		with tCommand as TMWFont do
        		SetFontInformation(Name,Size,Style);
          inc(SkipNext);
          tCommand := Page.Get(x+2);
          if tCommand.Command = commTab then
          begin
      			with tCommand as TMWTab do
        			SetTab(Tab);
          	tCommand := Page.Get(x+3);
            inc(SkipNext);
          end;
          with tCommand as TMWText do
          begin
						Buffer := Format( PageNumber.Text,[iPage] );
  					if PageNumber.ShowTotalPages then
  						Buffer := Buffer + ' of ' + IntToStr(TotalPages);
            WriteLine(X,Y,Buffer);
          end;
          inc(SkipNext);
        except
        	ShowMessage('Error writing page number information.');
				end;
      commPageNumber:
      	try
      	with tCommand as TMWPageNumber do
        	SetPageNumberInformation(yPosition,Text,Alignment,IncludeTotal,FontName,Size,Style);
        except
        	ShowMessage('Error Setting Page Number Information.');
        end;
      commTab:
      	try
      	with tCommand as TMWTab do
        	SetTab(Tab);
        except
        	ShowMessage('Error Setting Tab.');
        end;
      commTopOfPage:
      	try
      		SetTopOfPage;
        except
        	ShowMessage('Error Setting Top of Page.');
        end;
      commSetY:
      	try
      	with tCommand as TMWSetY do
        	SetYPosition(yPosition);
        except
        	ShowMessage('Error Setting Vertical Position.');
        end;
      commText:
      	try
      	with tCommand as TMWText do
        begin
        	oldColor := TheCanvas.Brush.Color;
        	TheCanvas.Brush.Color := BrushColor;
        	WriteLine(X,Y,Text);
          TheCanvas.Brush.Color := oldColor;
        end;
      	except
        	ShowMessage('Error Writing Line of Text.');
        end;
      commColumn:
      	try
      	with tCommand as TMWColumn do
        	CreateColumn(Number,XPosition,Length);
        except
        	ShowMessage('Error Storing Column Setting.');
        end;
    end
    end
    else
    	dec(SkipNext);
  end;
end;

procedure TPrintObject.UpdateBar(x: integer);
{ Sets the current progress bar value }
begin
	if UseProgress then
		ProgressBar.UpdateBar(x);
end;

procedure TPrintObject.IncBar(x: integer);
{ Increments the progress bar value by X }
begin
	if UseProgress then
		ProgressBar.IncBar(x);
end;

procedure TPrintObject.Increment;
{ Increment the progress bar with the value of ProgressInc }
begin
	if UseProgress then
  	ProgressBar.IncBar(ProgressInc);
end;

function TPrintObject.GetShowTotalPages: boolean;
{ Function to return the setting of showing total page numbers }
begin
	result := PageNumber.ShowTotalPages;
end;

procedure TPrintObject.SetShowTotalPages(Mode: boolean);
{ Sets whether total page numbers appears in page numbering }
begin
	PageNumber.ShowTotalPages := Mode
end;

procedure TPrintObject.OnPageChange;
begin
	if assigned(FPageChange) then FPageChange(self);
end;

procedure Register;
{ Register this component }
begin
  RegisterComponents('System', [TPrintObject]);
end;

procedure TPrintObject.Cleanup;
var
	x: integer;

begin
	for x := 1 to HeaderLines do
  	Header[x].Text := '';
  for x := 1 to FooterLines do
  	Footer[x].Text := '';
  for x := 1 to Columns do
  begin
  	ColumnInformation[x].XPosition := 0.0;
    ColumnInformation[x].Length := 0.0;
  end;
  PageNumber.Text := '';
  SetTab(0.0);
end;

end.

